//------------------------------------------------------------------------
//
// Booster_Timer
//
// Copyright (c) 2006 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      Booster_Timer.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2007-08-29 V0.1 start
//            2008-10-01 V0.2 modified for Booster2, no cutout yet.
//            2009-03-17 V0.3 added immediate try to ENABLE if dcc_state
//                            changes from 0 to 1.
//            
//------------------------------------------------------------------------
//
// purpose:   Intelligent Timer for turning off the Booster
//          
// content:    1. Defines and variable definitions
//             2. Init
//             3. ISR (only timerval++)
//             4. MAIN: check the states, trace and retry enable
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>

#define SIMULATION  0            // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine (toggle fast)
                                 // 3: test motor

//---------------------------------------------------------------------
// Timing Definitions: (all values in us)
//
#ifndef F_CPU
   // prevent compiler error by supplying a default 
   # warning "F_CPU not defined for <booster_timer.c>", set default to 8MHz 
   # define F_CPU 8000000UL
   // if changed: check every place where it is used
   // (possible range underflow/overflow in preprocessor!)
#endif

//---------------------------------------------------------------------
//
// Controlling the H-Bridge:
//
// a) DCC failure:
//    DCC must have a change every 12ms; if there is no change for 15ms,
//    turn off; turn on after a change (DCC state)
//
// b) Short detection:
//    after a short is detected for more than 100us (this time is done via HW)
//    turn off H-bridge.
//    try to turn on again: three times after 4ms
//    if this fails: wait 1s, then try again.
//
//---------------------------------------------------------------------
//
// Connect the Attiny2313 as follows:
// (outer pin mumbers: old 74123 Pinning)
//
//                      --------U--------
//                     -| 1 PA2   VCC 20|- ------------->>  
//   SHORT(in, low)    -| 2 PD0   PB7 19|- 
//   LOADED(in, low)   -| 3 PD1   PB6 18|-  
//                     -| 4 PA1   PB5 17|-  
//   LED orange(out)   -| 5 PA0   PB4 16|-
//   DCC(in, INT0)     -| 6 PD2   PB3 15|-  LED grn (active low)
//                     -| 7 PD3   PB2 14|-  JUMPER (in)
//   LED rot(out, low) -| 8 PD4   PB1 13|-  
//   ENABLE (out, hi)  -| 9 PD5   PB0 12|-  DATA_ON (out) (low=cutout)
//          |--------- -|10 GND   PD6 11|-
//                      -----------------
//
//
// This design must run with internal clock (XTAL1 is used as IO)
//
// JUMPER: if fitted: create CUTOUT
//
//---------------------------------------------------------------------
// Definition for OUTPUT and Timing

#define TICK_PERIOD  1000L               // 1ms tick for Timer

#define DCC_TIMEOUT_PERIOD      15000L       // 15ms

#define FAST_RECOVER_PERIOD      4000L

#define SLOW_RECOVER_PERIOD   1000000L 

#define BLINK_PERIOD           600000L
//---------------------------------------------------------------------------
// PORT Definitions:
//
// PORTB:
//
#define DATA_ON     0     // out, if low: create cutout  
#define NU_PB1      1     // in, unused
#define JUMPER      2     // in, internal pullup
#define LED_GREEN   3     // out, active low
#define NU_PB4      4     // in, unused
#define NU_PB5      5     // in, unused
#define NU_PB6      6     // in, unused
#define NU_PB7      7     // in, unused

// PORTD:

#define SHORT       0     // in: active low
#define LOADED      1     // in: active low
#define DCC_IN      2     // in, pullup
#define NU_PD3      3     // in, unused
#define LED_RED     4     // out, active low
#define ENABLE      5     // out: active high

// PORTA:

#define LED_ORANGE  0     // out, active low


#define DCC_IN_HIGH        ( (PIND & (1<<DCC_IN)))
#define DCC_IN_LOW         (!(PIND & (1<<DCC_IN)))

#define LED_GREEN_OFF       PORTB |= (1<<LED_GREEN)
#define LED_GREEN_ON        PORTB &= ~(1<<LED_GREEN)

#define LED_ORANGE_OFF      PORTA |= (1<<LED_ORANGE)
#define LED_ORANGE_ON       PORTA &= ~(1<<LED_ORANGE)

#define LED_RED_OFF         PORTD |= (1<<LED_RED)
#define LED_RED_ON          PORTD &= ~(1<<LED_RED)

#define IS_SHORT            (!(PIND & (1<<SHORT)))
#define IS_LOADED           (!(PIND & (1<<LOADED)))

#define JUMPER_FITTED       (!(PINB & (1<<JUMPER)))

#define ENABLE_OUT          PORTD |= (1<<ENABLE)
#define DISABLE_OUT         PORTD &= ~(1<<ENABLE)
#define ENABLED             (PORTD & (1<<ENABLE))        // as state var

#define NO_CUTOUT           PORTB |= (1<<DATA_ON)
#define MAKE_CUTOUT         PORTB &= ~(1<<DATA_ON)
#define IS_CUTOUT           (!(PORTD & (1<<ENABLE)))        // as state var



//
//----------------------------------------------------------------------------
// Global Data

unsigned char version[] PROGMEM    = {"-- Booster_V2 V1.02 "};
unsigned char copyright1[] PROGMEM = {"-- (c) Kufer Wolfgang, 17.03.2009 "};
unsigned char copyright2[] PROGMEM = {"-- www.opendcc.de --"};

void init_main(void)
  {
    DDRB  = (1 << DATA_ON )        // out, if low: create cutout  
          | (0 << NU_PB1  )         // in, unused
          | (0 << JUMPER   )        // in, internal pullup
          | (1 << LED_GREEN  )      // out, active low
          | (0 << NU_PB4 )          // in, unused
          | (0 << NU_PB5  )         // in, unused
          | (0 << NU_PB6  )         // in, unused
          | (0 << NU_PB7   );       // in, unused


    PORTB  = (1 << DATA_ON )        // out, if low: create cutout  
          | (0 << NU_PB1  )         // in, unused
          | (1 << JUMPER   )        // in, internal pullup
          | (1 << LED_GREEN  )      // out, active low
          | (0 << NU_PB4 )          // in, unused
          | (0 << NU_PB5  )         // in, unused
          | (0 << NU_PB6  )         // in, unused
          | (0 << NU_PB7   );       // in, unused
          

    DDRD  = (0<<  SHORT )       // in: active low
        | (0<<  LOADED  )          // in: active low
        | (0<<  DCC_IN  )          // in, pullup
        | (0<<  NU_PD3  )          // in, unused
        | (1<<  LED_RED )          // out, active low
        | (1<<  ENABLE  );         // out: active high
    
    
    PORTD  = (1<<  SHORT )       // in: active low -> pullup
        | (1<<  LOADED  )          // in: active low
        | (1<<  DCC_IN  )          // in, pullup
        | (0<<  NU_PD3  )          // in, unused
        | (0<<  LED_RED )          // out, active low
        | (0<<  ENABLE  );         // out: active high
    

    DDRA  = (1 << LED_ORANGE);        //    
    
    PORTA = (0 << LED_ORANGE);        

    // LEDs all on
    
    // Init Timer0 as CTC 
    // check PWM_TICK_PERIOD and F_CPU
    #define TIMER0_CLKDIV 64              // possible values: 1, 8, 64, 256, 1024

    #if (F_CPU / 1000000L * TICK_PERIOD / TIMER0_CLKDIV) > 255L
      #warning: overflow in OCR0A - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a larger clkdiv
    #endif    
    #if (F_CPU / 1000000L * TICK_PERIOD / TIMER0_CLKDIV) < 30L
      #warning: resolution accuracy in OCR0A too low - check TICK_PERIOD and F_CPU
      #warning: suggestion: use a smaller clkdiv
    #endif    


    OCR0A = F_CPU / 1000000L * TICK_PERIOD / TIMER0_CLKDIV ;  
    TCCR0A = (0 << COM0A1)      // compare match A
           | (0 << COM0A0) 
           | (0 << COM0B1)      // compare match B
           | (0 << COM0B0) 
           | 0                  // reserved
           | 0                  // reserved
           | (1 << WGM01)  
           | (0 << WGM00);      // Timer0 Mode 2 = CTC - Int on compare A
    TCCR0B = (0 << FOC0A) 
           | (0 << FOC0B) 
           | (0 << WGM02) 
    #if   TIMER0_CLKDIV == 1
           | (0 << CS02) | (0 << CS01)  | (1 << CS00);       // CS[2:0]=001 clkdiv 1
    #elif TIMER0_CLKDIV == 8
           | (0 << CS02) | (1 << CS01)  | (0 << CS00);       // CS[2:0]=010 clkdiv 8
    #elif TIMER0_CLKDIV == 64
           | (0 << CS02) | (1 << CS01)  | (1 << CS00);       // CS[2:0]=011 clkdiv 64 
    #elif TIMER0_CLKDIV == 256
           | (1 << CS02) | (0 << CS01)  | (0 << CS00);       // CS[2:0]=100 clkdiv 256
    #elif TIMER0_CLKDIV == 1024
           | (1 << CS02) | (0 << CS01)  | (1 << CS00);       // CS[2:0]=101 clkdiv 1024   
    #else
     #warning: TIMER0_CLKDIV is void
    #endif


    TIMSK = (0<<TOIE1)       // Timer1 Overflow
          | (1<<OCIE1A)      // Timer1 Compare A
          | (0<<OCIE1B)      // Timer1 Compare B
          | 0                // reserved
          | (0<<ICIE1)       // Timer1 Input Capture
          | (0<<OCIE0B)      // Timer0 Compare B
          | (0<<TOIE0)       // Timer0 Overflow
          | (1<<OCIE0A);     // Timer0 Compare A

    // Init Interrupt 0

    GIMSK = (0<<INT1)        // Int1 is msb
          | (0<<INT0)        // 1= Enable INT0 -> DCC
          | (0<<PCIE); 

    MCUCR = 0x03;  //           ;The rising edge of INT0 generates an interrupt request.  
  }


//==============================================================================
//
// Section 3
//


volatile signed int timerval;

signed int last_enable;
signed int last_short;
signed int last_loaded;
signed int last_dcc_change;

unsigned char dcc_level;
unsigned char dcc_state;

unsigned char fast_recover;


void blink_led(void)
  {
    volatile signed int led_time;
    unsigned char i;
    
    for (i=0; i<3; i++)
      {
        LED_RED_ON; LED_ORANGE_ON;  LED_GREEN_ON;
        led_time = timerval;
        while ((timerval - led_time) < (BLINK_PERIOD / 2 / TICK_PERIOD));
        LED_RED_OFF; LED_ORANGE_OFF; LED_GREEN_OFF;
        led_time = timerval;
        while ((timerval - led_time) < (BLINK_PERIOD / 2 / TICK_PERIOD));
      }    
  }


ISR(TIMER0_COMPA_vect)                      // Timer0 Compare Int
  {
    timerval++;
  }

// return true if a change in dcc-input is freshly detected
unsigned char check_dcc_change(void)
  {
    unsigned char retval = 0;
    if (dcc_level)
      {
        if (DCC_IN_LOW)
          {
            dcc_level = 0;
            cli();
            last_dcc_change = timerval;
            if (!dcc_state) retval=1;
            dcc_state = 1;                                      // DCC is okay
            LED_GREEN_ON;
            sei();
          }
      }
    else
      {
        if (DCC_IN_HIGH)
          {
            dcc_level = 1;
            cli();
            last_dcc_change = timerval;
            if (!dcc_state) retval=1;
            dcc_state = 1;
            LED_GREEN_ON;
            sei();
          }
      }
    return(retval);
  }


int main(void)
  {
    init_main();
     
    sei();

    blink_led();

    while(1)
      {
        if (check_dcc_change()) 
          {
            ENABLE_OUT; // try to enable Output if DCC coming back
            LED_RED_OFF;
            last_enable = timerval;
          }

        cli();
        if ((timerval - last_dcc_change) > (DCC_TIMEOUT_PERIOD / TICK_PERIOD))
          {
            dcc_state = 0;
            LED_GREEN_OFF;
            DISABLE_OUT;
          }
        sei();

        if (IS_LOADED) LED_ORANGE_ON;
        else LED_ORANGE_OFF;

        if (IS_SHORT)
          {
            if (ENABLED)
              {
                cli();
                last_short = timerval;
                sei();
              }
            DISABLE_OUT;
            LED_RED_ON;
          }
        else
          {
            // check, whether we are stable on - if so, set fast_recover allowed
            cli();
            if ((timerval - last_enable) > (2L * SLOW_RECOVER_PERIOD / TICK_PERIOD))
              {
                fast_recover = 3;
              }
            sei();
            
            if (!(ENABLED))
              { // try to enable output
                if (fast_recover > 0)
                  {
                    cli();
                    if ((timerval - last_short) > (FAST_RECOVER_PERIOD / TICK_PERIOD))
                      {
                        fast_recover--;
                        LED_RED_OFF;
                        last_short = timerval;   // shift reference point to actual
                        last_enable = timerval;
                        if (dcc_state)
                          {
                            ENABLE_OUT;
                          }
                      }
                    sei();
                  }
                else
                  {
                    cli();
                    if ((timerval - last_short) > (SLOW_RECOVER_PERIOD / TICK_PERIOD))
                      {
                        LED_RED_OFF;
                        last_short = timerval;   // shift reference point to actual
                        last_enable = timerval;
                        if (dcc_state)
                          {
                            ENABLE_OUT;
                          }
                      }
                    sei();
                  }
              }
          }
      }

    return(0);  // dummy  
  }
